#coding:utf-8
"""Модуль contr содержит классы-обертки для контроллера ТЭНа и контроллера
клапана отбора, подключенных к шине I2C микрокомпьютера Raspberry Pi
История изменений - в конце файла
OldBean, 17.01.27"""
import os
import smbus
import sens
import time
#-------------------------------------------------------------------------------
class TEH(object):
  """     Класс-обертка нагревателя куба."""
  def __new__(cls, vs = None, addrTEH = 0x04, W_220 = 1418.0, bus = smbus.SMBus(1)):
    """Перед созданием объекта в методе __new__ производится
       тестирование контроллера (наличие и работоспособность).
       vs - объект-датчик RMS
       addrTEH - адрес контроллера ТЭНа на шине I2C
       W_220 - номинальная мощность ТЭНа (при напряжении 220В)
       bus - объект-шина SMBus. По умолчанию - SMBus(1)."""
    try: # Проверим наличие и работоспособность контроллера ТЭНа
      bus.read_byte(addrTEH)
      # Все в порядке - создаем объект - контроллер ТЭНа
      return super(TEH, cls).__new__(cls) 
    except: 
      print "Контроллер ТЭНа отсутствует или неисправен"
      return None

  def __init__(self, vs = None, addrTEH = 0x04, W_220 = 1418.0, bus = smbus.SMBus(1)):
    """       vs - объект-датчик RMS
       addrTEH - адрес контроллера ТЭНа на шине I2C
       W_220 - номинальная мощность ТЭНа (при напряжении 220В)
       bus - объект-шина SMBus. По умолчанию - SMBus(1)."""
    self._addrTEH = addrTEH
    self._vs = vs
    self._W_220 = W_220
    self._coef = 4840000.0/self._W_220
    self._bus = bus
    if self._vs == None: # Нужно бы предупредить
      print """Предупреждение! Объект-датчик RMS не указан.
Устанавливаемая мощность ТЭНа может не соответствовать реальной!"""

  def __del__(self): # В деструкторе выключим ТЭН еще раз на всякий случай
    self._bus.write_byte(self._addrTEH, 0)

  @property
  def W(self):
    """       Свойство - текущая мощность ТЭНа в ваттах.
       При каждом использовании этого атрибута в правой части выражения
       производится чтение данных из контроллера ТЭНа."""
    teh_pc = self._bus.read_byte(self._addrTEH)*0.01
    if self._vs == None: # Работаем без датчика RMS
      wrk = 1.0
    else:
      wrk = self._vs.V*0.00454545454545 # V/220
    return teh_pc*self._W_220*wrk*wrk

  @W.setter
  def W(self, val): # val - желаемая мощность ТЭНа в ваттах
    """       Свойство - устанавливаемая мощность ТЭНа в ваттах.
       Если датчик RMS присутствует в системе, то мощность устанавливается
       с учетом текущего действующего значения напряжения сети."""
    if self._vs == None: v = 220.0 # Работаем без датчика RMS
    else: v = self._vs.V

    if v < 140.0:
      print "Напряжение в сети < %3d В. Не буду работать." % (v)
      self._bus.write_byte(self._addrTEH, 0)
      return

    if val < 0: wpc = 0
    elif val > self._W_220: wpc = 100
    else:
      wrk = 220.0/v
      wpc = int(((100.0*val)/self._W_220)*wrk*wrk + 0.5)
      if wpc > 100: wpc = 100
      
    # Попытаемся записать байт в ведомое устройство по шине I2C
    tries = 10; error = None
    while tries: # Если будет сбой при записи - сделаем tries попыток
      try: self._bus.write_byte(self._addrTEH, wpc)
      except IOError as e:
        error = e; tries -= 1
        print "tries = " + str(tries) + "; " + str(error)
      else: break # Нормально отправили байт - выходим из цикла
    if not tries: raise error # Сбой устойчивый - бросаем исключение и 
                              # "завязываем" с этим делом
#-------------------------------------------------------------------------------
class SD(object): # Класс устройства отбора
  """Класс-обертка для контроллера клапана отбора"""

  def __new__(cls, addr = 0x06, coef = 9.5053486617, bus = smbus.SMBus(1)):
    """Перед созданием объекта в методе __new__ производится
       тестирование контроллера (наличие и работоспособность).
       addr - адрес контроллера на шине I2C
       coef - коэффициент для пересчета % ШИМ в мл/час. Значение этого
              параметра должно быть измерено в процессе калибровки уже
              смонтированного на установке клапана отбора. Определяется
              при помощи линейной регрессии графика зависимости скорости
              отбора клапана от % ШИМ (на спирте).
       bus - объект SMBus - шина I2C."""
    try: # Проверим наличие и работоспособность контроллера
      bus.read_byte(addr)
      # Все в порядке - создаем объект - контроллер клапана отбора
      return super(SD, cls).__new__(cls)
    except:
      print "Контроллер клапана отбора отсутствует или неисправен"
      return None

  def __init__(self, addr = 0x06, coef = 9.5053486617, bus = smbus.SMBus(1)):
    """       addr - адрес контроллера на шине I2C
       coef - коэффициент для пересчета % ШИМ в мл/час. Значение этого
              параметра должно быть измерено в процессе калибровки уже
              смонтированного на установке клапана отбора. Определяется
              при помощи линейной регрессии графика зависимости скорости
              отбора клапана от % ШИМ (на спирте).
       bus - объект SMBus - шина I2C."""
    self._addr = addr
    self._coef = coef
    self._coefi = 1.0/coef # Для обратного пересчета (мл/час в % ШИМ)
    self._bus = bus
    self._bus.write_byte(self._addr, 0)
    self._val = 0.0 # Задаваемый расход в мл/час
    self._flow = 0.0 # Суммарный расход после последнего сброса
    self._start = time.time() # Начало установки очередной скорости отбора

  def __del__(self): # В деструкторе выключим клапан еще раз на всякий случай
    self._bus.write_byte(self._addr, 0)

  @property
  def Q(self):
    """       Свойство - текущее значение скорости отбора в мл/час.
       Значение ШИМ контроллера считывается каждый раз, когда этот
       атрибут появляется в правой части выражения"""
    return self._bus.read_byte(self._addr)*self._coef

  @Q.setter
  def Q(self, val): # val - расход в мл/час
    """       Свойство - задаваемое значение скорости отбора в мл/час.
       Необходимый уровень ШИМ будет устанавливаться на контроллере
       клапана отбора каждый раз, когда этот атрибут используется
       в левой части выражения перед знаком присваивания."""
    if val < 0: qpc = 0
    else: qpc = int(val*self._coefi + 0.5) # qpc - % ШИМ
    if qpc > 100: qpc = 100 # Ограничение на 100%
      
    # Попытаемся записать байт в ведомое устройство по шине I2C
    tries = 10; error = None
    while tries: # Если будет сбой при записи - сделаем tries попыток
      try: self._bus.write_byte(self._addr, qpc)
      except IOError as e:
        error = e; tries -= 1
        print "tries = " + str(tries) + "; " + str(error)
      else: break # Нормально отправили байт - выходим из цикла
    if not tries: raise error # Сбой устойчивый - бросаем исключение и 
                              # завязываем с этим делом
    now = time.time()
    """Скорость отбора, возможно, изменилась. Поэтому зафиксируем расход
    на текущий момент, прибавим его к счетчику расхода  и начнем отсчет
    нового периода времени"""
    self._flow += self._val*(now - self._start)*2.7777778e-4 # сек -> час
    self._val = val
    self._start = now

  def FZ(self):
    """Сброс счетчика расхода"""
    self._flow = 0.0
    self._start = time.time()

  @property
  def F(self):
    """Свойство - общий расход с момента последнего сброса (вызов метода FZ()
       или создания объекта) счетчика расхода (self._flow)"""
    now = time.time() # Фиксируем текущий момент времени
    """Далее прибавляем к счетчику расхода (self._flow) расход, накопившийся
       с последнего обращения к свойству F или с момента последнего изменения
       расхода (в сеттере Q). Этот момент зафиксирован в переменной self._start"""
    self._flow += self._val*(now - self._start)*2.7777778e-4 # сек -> час
    self._start = now # Обновим начало очередного промежутка времени
    return self._flow
#-------------------------------------------------------------------------------
if __name__ == "__main__": # Тестирование классов модуля sens
  print("\nКонтроллер ТЭНа")
  vs = sens.RMS() # Сначала создадим объект-датчик RMS
  teh = TEH(vs) # Теперь попытаемся содать объект-контроллер ТЭНа
  if teh != None: # Объект-контроллер ТЭНа удачно создан
    print "%5.1f Вт" % (teh.W)
    teh.W = 600 # Установим мощность нагрева 600 Вт
    print "%5.1f Вт - рабочий режим" % (teh.W) # Проверим так ли это
    time.sleep(10) # Немножко подремлем...
    teh.W = 2000 # Теперь мощность на максимум (пожелаем мощность заведомо
                 # бОльшую, чем номинальная мощность ТЭНа)
    print "%5.1f Вт - максимальная мощность" % (teh.W) # Посмотрим на мощность
    time.sleep(10)
    teh.W = 0 # Выключим ТЭН
    print "%5.1f Вт - выключен" % (teh.W) # Проверим на всякий случай

  print("\nКонтроллер ТЭНа. Работа без датчика RMS")
  teh = TEH() # Датчик RMS по умолчанию отсутствует
  if teh != None: # Объект-контроллер ТЭНа удачно создан
    print "%5.1f Вт" % (teh.W)
    teh.W = 600
    print "%5.1f Вт - рабочий режим" % (teh.W)
    time.sleep(10)
    teh.W = 2000
    print "%5.1f Вт - максимальная мощность" % (teh.W)
    time.sleep(10)
    teh.W = 0
    print "%5.1f Вт - выключен" % (teh.W)

  print("\nУстройство отбора")
  sd = SD() # Попытаемся создать объект-контроллер клапана отбора
  if sd != None: # Все тип-топ
    print "%6.1f мл/час" % (sd.Q)
    sd.Q = 400.0 # Установим скорость отбора 400 мл/час
    print "%6.1f мл/час - рабочий режим" % (sd.Q) # Проверим 
    time.sleep(20) # Пауза - клапан должен щелкнуть два раза
    sd.Q = 0.0 # Выключим отбор
    print "%6.1f мл/час - выключен" % (sd.Q) # Опять проверим
    print "Оценка расход за время теста ~ %5.1f мл" % (sd.F)
#---------------------------------------------------------------------------------------
"""История изменений

19.01.2017
----------
1. В сеттеры классов TEH и SD добавлена обработка исключения IOError 
   при посылке байта соотвествующим контроллерам. При возникновении
   исключения будут сделаны еще попытки (всего до 10) отправить байт.
   Если сбой устойчивый (все 10 попыток неудачны) - исключение кидается
   выше. 

25.01.2017 
----------
2. Расширена документация модуля
3. В деструкторы добалено выключение соответствующих устройств

26.01.2017
----------
4. Добавлена проверка наличия и работоспособности контроллеров до
   создания объекта. Если проверка неудачна, то переменной объекта
   присваивается None
5. Расширен функционал контроллера клапана отбора - добавлен учет расхода
   продукта через клапан отбора: свойство F - суммарный расход с момента
   создания объекта-контроллера клапана отбора или с момента последнего
   сброса и собственно метод FZ() - "сброс".
"""
