#coding:utf-8
import sys, os, time
import math, collections
import numpy as np
import sens, contr, kbh, misc

""" Файл nna_36.py
    1. Устранение "дребезга" температуры, старт-стопе без гистерезиса
    2. Уставка - по локальному (за время текущего режима) максимуму Ts[cid]
    3. Стоп срабатывает когда температура в колонне Ts[cid] превышает или равна
       уставке И ее производная положительна
    4. Старт (во время паузы, т.е. после стопа) происходит тогда, когда 
       производная Ts[cid] становится <= 0 ИЛИ если сама температура в колонне
       становится меньше уставки.
    5. Фиксировано время выполнения основного цикла

    OldBean, 25.04.2017"""
#-------------------------------------------------------------------------------
# Параметры установки и автоматики
#-------------------------------------------------------------------------------
hid = 0x180a353 # ID датчика температуры куба
cid = 0x18c5c68 # ID датчика температуры в нижней части колонны
did = 0x18c6520 # ID датчика температуры на входе дефлегматора
wid = 0x180c746 # ID датчика температуры выходящей воды охлаждения дефлегматора

""" tids - список идентификаторов датчиков. Порядок датчиков в списке  важен,
    т.к. их номер в списке (0-2) используется в карте режимов:
    0 - датчик в кубе, 1 - датчик в колонне, 2 - датчик в дефлегматоре """
tids = (hid, cid, did, wid) 
print '-'*80

qT      = 0.0625    # Квант датчика DS18B20
qT05    = 0.03125   # 0.5*qT
TwidMax = 45.0      # Максимальная температуры воды охлаждения дефлегматора

""" Карта режимов работы установки. Каждый режим представлен именованным 
 кортежем, содержащим следующие параметры:
   W     - мощность нагрева ТЭНа, Вт
   stab  - необходимость стабилизации мощности нагрева ТЭНа, True/False
   Q     - номинальная скорость отбора, мл/час
   reg   - возможность регулирования скорости отбора
           <= 0 - регулирование отбора не производится
           1 - регулирование методом "старт-стоп". Если вес wt равен 1.0,
               то обычный старт-стоп с фиксированным уровнем скорости
               отбора. Если wt < 1, то базовый уровень корректируется с учетом
               температуры в кубе (вес предыдущей скорости отбора - wt, а
               а рассчитанной по температуре в кубе - (1.0 - wt)
           2 - отбор по температуре в кубе с фиксированным флегмовым числом,
               определенным в момент активирования регулятора (клавиша 's')
           ...
   Fmax  - окончание режима при достижении данного суммарного расхода.
           Если -1, то нет контроля
   tmax  - окончание режима по времени (мин). Если -1, то нет контроля
   tsn   - номер датчика температуры, по которому осуществляется контроль:
           по превышению температуры происходит смена режима. Если -1,
           то контроля нет. Датчики нумеруются следующим образом:
           0 - в кубе, 1 - в колонне, 2 - в дефлегматоре
   T     - Порог температуры, после которого происходит смена режима
   nextm - номер следующего режима, в который система должна перейти по
           окончанию текущего
 Для изменения параметров режимов, измените соответствующие значения кортежей
"""
mrec = collections.namedtuple('mrec', 'W stab Q reg Fmax tmax tsn T nextm')
modes = [] # Список режимов
#                   W    stab     Q  reg  Fmax tmax  tsn    T nextr
modes.append(mrec(   0, False,    0, -1,   -1,  -1,  -1,   -1, 0))  # 0 - Мониторинг
modes.append(mrec(2000, False,    0, -1,   -1,  -1,   1, 60.0, 2))  # 1 - Разгон
modes.append(mrec( 600,  True,    0, -1,   -1,  -1,  -1,   -1, 2))  # 2 - Холостой ход
modes.append(mrec( 600,  True,   50, -1,  200,  -1,  -1,   -1, 2))  # 3 - Головы
modes.append(mrec( 600,  True,  500,  1, 1000,  -1,  -1,   -1, 2))  # 4 - Подголовники
modes.append(mrec( 600,  True,  500,  1, 7500,  -1,   0, 98.5, 2))  # 5 - Тело
modes.append(mrec( 600,  True,  500, -1, 1000,  -1,   2, 78.5, 2))  # 6 - Хвосты
modes.append(mrec(2000, False, 1000, -1,   -1,  60,  -1,   -1, 2))  # 7 - Пропарка колонны
modes.append(mrec(   0, False,    0, -1,   -1,  -1,  -1,   -1, 2))  # 8 - резерв
modes.append(mrec(   0, False,    0, -1,   -1,  -1,  -1,   -1, 2))  # 9 - резерв

print "Карта режимов (N, W, stab, Q, reg, Fmax, tmax, tsn, T, nextm):"
for i in range(len(modes)):
  print i, '\t',
  for par in modes[i]:
    print par, '\t',
  print
print '-'*80

""" wt - вес старого базового уровня скорости отбора, при расчете нового: 
    bQ = bQ*wt + cQ*(1-wt), где cQ - уровень, рассчитанный по температуре
    в кубе """
wt = 0.8
print "Весовой коэффициент регулятора старт-стоп wt = %3.2f" % (wt)
print '-'*80

Qmin    = 0.0       # Минимальная скорость отбора (для регулятора)
Qmax    = 900.0     # Максимальная скорость отбора (для регулятора)
#-------------------------------------------------------------------------------
# Создание и тестирование объектов датчиков и контроллеров
#-------------------------------------------------------------------------------
vs = sens.RMS()     # Создаем объект - датчик RMS
ps = sens.BMP180()  # Создаем объект - датчик атмосферного давления
teh = contr.TEH(vs) # Создадим объект "ТЭН"
sd = contr.SD()     # Создадим объект "устройство отбора"
if not misc.test(tids, vs, ps, teh, sd): # Проверяем работоспособность "железа"
  print "Некоторые элементы оборудования отсутствуют или неисправны. Завершение работы."
  sys.exit(1)
#-------------------------------------------------------------------------------
# Инициализация других переменных и подготовка к работе
#-------------------------------------------------------------------------------
timeout = 1000            
""" Переведем консоль в неблокирующий режим работы. Это необходимо для эмуляции
    функции kbhit() """
tt = kbh.TT(timeout)      

""" Открываем файл журнала (log) и определяем начало локального веремени (tst)
    и начало текущего режима (rst)"""
(log, tst, rst) = misc.logOpen("log")

mode = old_mode = 0 # В начале работы система работает в режиме 0

regFlag = False     # Сбрасываем индикатор активности регулятора отбора
curQ = 0.0          # Текущая скорость отбора
Tset = -273.15      # Текущее значение уставки
cR = 0.0            # Коэффициент для регулятора по температуре куба
bQ = 0.0            # Текущий базовый уровень скорости отбора
stopFlag = False    # Индикатор паузы (стопа) в отборе

ctrp = 0            # Счетчик актов выполнения условия для "стопа" подряд. 
                    # Это необходимо для устранения "дребезга" параметров, 
                    # обусловленного дискретным характером цифровых датчиков
ctrm = 0            # То же, только для "старта"
TcMax = 0.0         # Макс. т-ра в колонне с начала текущего режима (точнее - 
                    # со смены типа регулятора (параметр modes[mode].reg)
#-------------------------------------------------------------------------------
al = 25             # Длина массивов для вычисления скользящих средних и т.п.
TcPrev = 0.0        # Предыдущее значение температуры в колонне
tPrev = -1.0        # Предыдущей момент времени
aderiv = np.zeros(al, dtype = np.float32)  # Массив значений производных
ip = 0              # Точка вставки в массив текущего значения температуры
maTcDeriv = 0.0     # Скользящее среднее производной температуры в колонне
#-------------------------------------------------------------------------------
# Подготовка закончена - входим в главный цикл приложения
#-------------------------------------------------------------------------------
print '-'*80
while True:
  # Измерение текущих параметров................................................
  now = time.time();     # Время
  tle = now + 4.8 # 4.8 секунд на цикл (0.08 мин)

  Ts = sens.DS18B20.Ts() # Температуры
  V = vs.V               # Действующее напряжение сети
  cW = teh.W             # Мощность нагрева
  P = ps.P               # Атмосферное давление         
  Q = sd.Q               # Скорость отбора
  F = sd.F               # Расход

  cTc = Ts[cid]
  if cTc > TcMax: TcMax = cTc # корр. текущий максимум т-ры в колонне
  if tPrev > 0: # Считаем производные и заносим их в кольцевой массив
    aderiv[ip] = 60.0*(cTc - TcPrev)/(now - tPrev) # В °C/мин
    ip += 1
    if ip == al: ip = 0
    maTcDeriv = np.average(aderiv) # Скользящее среднее производной т-ры
  TcPrev = cTc
  tPrev = now

  # Обрабатываем состояние клавиатуры...........................................
  if tt.kbhit(): # Какая-то клавиша была нажата
    ch = tt.getch()
    if ch == 'q': break # Выходим из цикла - конец работы
    elif ch == 's' and modes[mode].reg > 0: # Активация регулятора отбора
      Tset0 = Tset = TcMax + qT05 # Запоминаем базовую уставку
      Pa0 = P                     # Запомним текущее атмосферное давление
      bQ = modes[mode].Q          # Начальный базовый уровень скорости отбора
      cR = bQ/(100.0 - Ts[hid])   # Коэффициент для расчета скорости отбора по
                                  # температуре куба
      regFlag = True              # Активируем регулятор скорости отбора
    elif ch == 'r' and modes[mode].reg > 0: # Выключение регулирования отбора
      regFlag = False    # Сбрасываем флаг активации регулятора
    elif ord(ch) >= 48 and ord(ch) <= 57: # Ввели номер режима
      old_mode = mode; mode = int(ch)
  
  # Протоколирование процесса...................................................
  str = " %8.4f" % (maTcDeriv)
  if regFlag:
    str += " %7.3f %4d" % (Tset, bQ)
  misc.logWrite(log, now, tst, rst, mode, tids, Ts, cW, P, Q, F, regFlag, str)
  misc.prnt(now, tst, rst, mode, tids, Ts, cW, P, Q, F, regFlag, str)

  # Переключаем режимы в зависимости от условий.................................
  if old_mode == mode: # Если режим не был изменен вручную
    if modes[mode].tsn >= 0:  # Есть контроль температуры с tsn-го датчика
      if Ts[tids[modes[mode].tsn]] >= modes[mode].T: # т-ра превышена - меняем режим
        old__mode = mode; mode = modes[old_mode].nextm

    if modes[mode].Fmax > 0: # Если есть ограничение на расход
      if F >= modes[mode].Fmax: # и он превышен - меняем режим
        old__mode = mode; mode = modes[old_mode].nextm
    
    if modes[mode].tmax > 0: # Если есть контроль на длительность режима 
      if (now - rst)/60.0 >= modes[mode].tmax: # и она превышена - меняем режим
        old__mode = mode; mode = modes[old_mode].nextm

  if Ts[wid] >= TwidMax: # Блокировка: если охлаждение дефлегматора недостаточно
    teh.W = 0; sd.Q = 0  # Отключаем нагрев и отбор
    old__mode = mode; mode = 0 # Переходим в режим мониторинга

  # Отрабатываем смену режима...................................................
  if old_mode != mode: 
    if not modes[mode].stab and mode != 0: # Режим без стабилизации мощности -
      teh.W = modes[mode].W # устанавливаем один раз в начале режима
    sd.Q = modes[mode].Q # Это же и для скорости отбора
    sd.FZ() # Сброс счетчика расхода при смене режима
    if modes[old_mode].reg != modes[mode].reg: # Дезактивируем регулятор
      regFlag = False
      TcMax = 0.0 # Сбрасываем максимум температуры в колонне
    rst = now # Фиксируем момент начала текущего режима
    old_mode = mode # Ну теперь вроде бы все сделали для смены режима

  """Для режимов со стабилизацией мощности, мощность обновляется на каждом
     такте главного цикла приложения (с учетом напряжения сети)"""
  if modes[mode].stab: teh.W = modes[mode].W

  if regFlag: # Регулятор активирован...........................................
    Tset = Tset0 + 0.033*(P - Pa0) # Корректируем уставку по давлению
    if modes[mode].reg == 2:       # Метод отбора по температуре в кубе с 
      bQ = cR*(100.0 - Ts[hid])    # фиксированным флегмовым числом

    """ Для устранения "дребезга" температуры, обусловленного дискретностью
        цифрового датчика температуры, производится подсчет количества
        превышений порога, возникающих подряд. Если, после превышения порога, на
        следующем шаге превышения нет (т.е. налицо "дребезг"), то предыдущее
        превышение нивелируется. Стоп наступает только после
        многократного (более двух раз подряд) превышения порога """
    if not stopFlag: # Отбор производится
      if cTc >= Tset - qT and  maTcDeriv > 0 : ctrp += 1 # Продвигаемся к стопу ;)
      elif ctrp > 0: ctrp -= 1     # Увы, это - "дребезг" температуры
      if ctrp > 2: # Порог превышен более 2 раз подряд, значит все - "стоп"
        curQ = Qmin
        stopFlag = True
        ctrp = 0 # Сбросим счетчик превышений
      else:
        curQ = bQ

    else: # Сейчас - пауза в отборе (т.е. был стоп)
      if cTc < Tset + qT and maTcDeriv <= 0 : ctrm += 1 # Продвигаемся с старту ;)
      elif ctrm > 0: ctrm -= 1 # Увы, это "дребезг" температуры
      if ctrm > 2: # Отриц. произв-ная была более 2 раз подряд - значит "старт"
        if modes[mode].reg == 1:
          """ Корректируем базовую скорость отбора с учетом веса wt старого
              значения (bQ) и рассчитанного по температуре в кубе (cQ) """
          cQ = cR*(100.0 - Ts[hid])
          bQ = bQ*wt + cQ*(1.0 - wt)
        curQ = bQ   # Старт
        stopFlag = False
        ctrm = 0 # Сброс счетчика

    sd.Q = curQ 

  # Задержка для выравнивания времени выполнения цикла
  while time.time() < tle: pass
#-------------------------------------------------------------------------------
log.close()
